/* Placed in the Public Domain 5/25/97  John E. Carter                   */
/* johnecarter@mindspring.com                                            */
/* Requires the Mark Williams C Compiler version 4.0.10, or appropriate  */
/* changes to work with your compiler.                                   */
/*                                                                       */
/* JREST - John's RESTore program                                        */
/* Restores individual files from a disk created by the MSDOS 'backup'   */
/* command of DOS 2.11 or 3.2.  Compiles under Mark Williams Let's C     */
/* Version 4.0.10.  It's MSDOS specific, but the code is fairly vanilla  */
/* and should compile under most MSDOS C compilers.                      */
/* See the comments for items that will need changes based on whether    */
/* your C compiler has mkdir(), find_first(), find_next() and some       */
/* other functions.  Work-arounds for the Mark Williams' compiler are    */
/* provided, others are "left as an exercise for the reader".            */
/*                                                                       */
/* Copyright 1989 - John E. Carter                                       */
/*                                                                       */
/* lines beginning with '<e>' are used by the Source Print program       */
/* <e> getstamp, putstamp, readkybd, getattrib, setattrib, getdta        */
/* <e> printf, fprintf, mkdir, regtop, ptoreg                            */

#include <stdio.h>
#include <dos.h>
#include <dosfind.h>
#include <ctype.h>

#define HDRSIZE    128
#define SECTORSIZE 512
/* This was orgiginally written for a PC with a 360K 5.25" floppy. */
/* However, it reads 1.44M 3.5" floppies faster than most programs. */
/* A larger buffer here might make a difference in the floppy read speed. */
#define TRACKSIZE  9216    /* 9 sectors/track * two sides */
#define MINUS1     -1
#define BIGMAGIC   0xFFFF

#define myname     "jrest"
#define BACKSLASH       92
#define int21      0x21

struct fcb_t jfcb, *fcbp;

struct reg *sregp, *dregp, sreg, dreg;

static char sccs_id[]="@(#) jrest.c - 1.13 - 12/12/89 johnecarter@mindspring.com";

main(argc,argv)
int argc;
char *argv[];
{
    char header[129], drive[3], backid[15], cmd[60], dest[3], filename[20];
    char dir[60], file[15], fullpath[80], command[80], oldpath[80];
    char tmp1[80], tmp2[80], *ptmp1, *ptmp2, *base, *index(), *rindex();
    char tmpname[20], ans[15], tmp[TRACKSIZE], overide[3];
    int c, i, s, first_slash, next_slash, last_slash, dest_def, overwrite;
    int last_part, int_part, complete, count, pos, one_part, outopen, over;
    int fpin, fpout, readin, rx, exists, isready, disknum, thisnum, getbknum();
    int attrib, erval, defdrv, dkc, shortfile;
    unsigned int dxdate, cxtime;
    long filesize;

#ifdef DEBUG
    for(i = 0;i < argc;i++)
        printf("argv[%d]: %s\n",i,argv[i]);
#endif

    /* point structure pointers to structures used */
    sregp = &sreg;
    dregp = &dreg;

/* init all the string arrays */
    for(i = 0;i < 129;i++)
        header[i] = '\0';
    for(i = 0;i < 60;i++)
    {
        cmd[i] = '\0';
        dir[i] = '\0';
    }
    for(i = 0;i < 15;i++)
    {
        ans[i] = '\0';
        file[i] = '\0';
        backid[i] = '\0';
    }
    for(i = 0; i < 80;i++)
    {
        fullpath[i] = '\0';
        oldpath[i] = '\0';
        tmp1[i] = '\0';
        tmp2[i] = '\0';
        command[i] = '\0';
    }

    /* init all the variables */
    erval = first_slash = next_slash = last_slash = complete = count = pos = 0;
    attrib = last_part = int_part = outopen = dest_def = overwrite = over = 0;
    dxdate = cxtime = fpin = fpout = rx = exists = isready = thisnum = 0;
    dkc = defdrv = shortfile = 0;
    filesize = 0L;

/* check for correct number of arguments */
    if(argc < 2)
    {
        usage();
        exit(2);
    }

    /* set default drive names */
    strcpy(drive,"A:");
    strcpy(dest,"C:");

    /* check for drive name on command line */
    if(index(argv[1],':') > 0)
    {
        for(i=0;i<2;i++)
            drive[i] = toupper(argv[1][i]);
        strcpy(filename,argv[1]);
    }        
    else
    {
        strcpy(dir,drive);
        strcat(dir,argv[1]);
        strcpy(filename,dir);
    }

    /* parse options - need code to handle filename if not first */
    for(i = 2;i < argc;i++)
    {
        if(argv[i][0] == '-' || argv[i][0] == '/')
        {
            switch(argv[i][1])
            {
                case 'd': i++;
                          strcpy(dest,argv[i]);
                          dest_def++;
                          dest[0] = toupper(dest[0]);
                          break;
                case 'o': overwrite++;
                          over++;
                          break;
                case 'r': isready++;
                          break;
                 default: fprintf(stderr,"\nUnknown option %s\n",argv[i]);
                          usage();
                          exit(4);
            }
        }
    }
        /* tell the user what we're doing */
    fprintf(stderr,"\nrestoring from drive %c: to drive %c:\n",drive[0],dest[0]);

/*****************************************************************************/
/* there are several sections with notes for future additions - have at it ! */
/*****************************************************************************/

/* check for valid source and destination drives */
/* complain and exit if no such drive */
/* valid() */

/* check default drive - if same as destination, skip 'source disk' message */
   if(!strcmp(drive,dest))
        defdrv++;

/* to get # of DOS default drive (if needed) */
/********************************* 
    get_default(sregp);
    intcall(sregp,dregp,int21);
    defdrv = dregp->r_ax & 0x0f;
********************************/

    if(!isready && !defdrv)
    {
        fprintf(stderr,"\nPress return when source disk in drive %c: is ready: ",drive[0]);
        c = getchar();
        isready++;
    }

    /* put drive and name of backup marker file in string */
    strcpy(backid,drive);
    strcat(backid,"BACKUPID.@@@");

  /* main program loop - this is where the work gets done */
  while(strcmp(filename,"NOMORE"))
  {
    c = 254;
    while((c = getbknum(backid,drive)) == 254) /* get backup disk number */
        ; /* null loop body */
        if(c == 0 && dkc != 0)
          c = dkc;
        fprintf(stderr,"\nUsing backup disk number %d\n\n",c);
        disknum = c;

        /* check that source file exists */
    if((fpin = open(filename,0)) == MINUS1)
    {
        fprintf(stderr,"\n%s: %s unreadable or does not exist\n",myname,filename);
        exit(1);
    }

    /* note that these are ONLY for non-hidden files */
    /* get date/time of backup file */
    getstamp(sregp,fpin);
    intcall(sregp,dregp,int21);
    dxdate = dregp->r_dx;
    cxtime = dregp->r_cx;

    /* get attributes of backup file */
    getattrib(sregp,filename);
    intcall(sregp, dregp, int21);
    attrib = dregp->r_cx;

    /* get size of backup file */
/********************************
    setdta(sregp,fcbp);
    intcall(sregp,dregp,int21);
********************************/

/********************************
    getdta(sregp);
    intcall(sregp,dregp,int21);
    regtop(dregp->r_bx,dregp->r_es,fcbp);
#ifdef DEBUG
    fprintf(stderr,"fcbp=%ld, %u, %d\n",fcbp,fcbp,fcbp);
#endif
    filesize = fcbp->f_size;
********************************/

    /* verify file exists */
/****************************************************************
    findmatch(sregp,filename);
    intcall(sregp,dregp,int21);
    if(dregp->r_ax == 0)    
        fprintf(stderr,"%s is %ld bytes\n",filename,filesize);
    else
        fprintf(stderr,"file %s not found\n",filename);
****************************************************************/


    readin = read(fpin,header,128);
    if(readin != 128)
    {
        fprintf(stderr,"%s is not a backup file\n",filename);
        close(fpin);
        exit(3);
    }
    header[128] = '\0';


     /* define to turn off check for backslash in file name (DOS 1.x) */
#ifndef NO_BKCK
    if(header[5] != BACKSLASH)
    {
        fprintf(stderr,"\n%s is not a backup file\n",filename);
        exit(3);
    }
#endif

#ifdef DEBUG
    fprintf(stderr,"\nDEBUG:\n");
    for(i = 0; i < 128 && isprint(header[i]);i++)
        putchar(header[i]);
    putchar('\n');
#endif

  /* header byte indicates intermediate segment (file continues on next */
  /* backup disk) or last segment (file ends on this disk)              */
    if(header[0] == 0)
        {
            int_part++;
            one_part = 0;
#ifdef VERBOSE
            if(header[1] != 0)
               fprintf(stderr,"working on part %d of multi-disk file\n",header[1]);
#endif
        }        
        else
        {
            last_part++;
            one_part++;
#ifdef VERBOSE
            if(header[1] != 0)
               fprintf(stderr,"working on part %d (last part) of multi-disk file\n",header[1]);
#endif
        }

    /* get full pathname of file from header, starting in byte 5 */
    for(i = 5,s = 0;header[i] != 0;i++,s++)
        fullpath[s] = header[i];
    fullpath[s] = '\0';

#ifdef DEBUG
    fprintf(stderr,"DEBUG: ");
    fprintf(stderr,"s=%d, i=%d\n",s,i);
    fprintf(stderr,"fullpath: size=%s, content=%s\n",strlen(fullpath),fullpath);
#endif

    if(!outopen)  /* if output file is not open */
    {
        /* parse the full pathname, key on \ as path separator       */
        /* make directories as needed to insure complete path exists */
        /* re-creating an existing directory returns "Access Denied" */
        /* (error 5) but does not affect the existing directory      */

        base = fullpath;  /* get address of array element 0 */
        last_slash = rindex(fullpath,BACKSLASH) - base; /* find last '\' */
        if(last_slash < 0)  /* if no '\' */
            last_slash = 0;

#ifdef DEBUG
        fprintf(stderr,"DEBUG: last_slash=%d\n",last_slash);
#endif

        exists = 0;

    /* define NOMKDIR if your compiler does not have mkdir() */
    /* or you can't translate the Let's C interrupt call to your compiler */

#ifdef NOMKDIR
        fprintf(stderr,"Checking directories - ");
        fprintf(stderr,"ignore 'Unable to create directory' messages\n");
#endif

        if(last_slash > 0)  /* if at least one '\' */
        {
            for(i = 0;i < last_slash;i++)  /* get path to last '\' */
                dir[i] = fullpath[i];
        }
        dir[i] = '\0';

        base = dir;
        first_slash = index(dir,BACKSLASH) - dir;  /* find first '\' */
        if(first_slash < 0)  /* if no '\' */
            first_slash = 0;

#ifdef DEBUG
    fprintf(stderr,"DEBUG: ");
        fprintf(stderr,"dir: size=%d, content=%s\n",strlen(dir),dir);
        fprintf(stderr,"first_slash=%d\n",first_slash);
#endif

        if(first_slash >= 0)  /* if at least one '\' */
        {
            ptmp1 = fullpath;
            base = ptmp1;
            ptmp1++;
            next_slash = index(ptmp1,BACKSLASH) - base; /* find next '\' */
#ifdef DEBUG
    fprintf(stderr,"next_slash=%d, base=%d, ptmp1=%d\n",next_slash,base,ptmp1);
#endif
            if(next_slash < 0)  /* if no next '\' */
                next_slash = 0;

            if(next_slash >= 0)  /* if at least one '\' */
            {                    /* parse each dir name and mkdir */
                complete = 0;
                strcpy(tmp1,fullpath);
                    ptmp1 = tmp1;
                    ptmp2 = tmp1;
                while(!complete)
                {
                    base = ptmp1;
                    ptmp1++;
                    pos = index(ptmp1,BACKSLASH) - base;
                    pos--;
#ifdef DEBUG
        fprintf(stderr,"pos=%d, ptmp1=%d\n",pos,ptmp1);
#endif
                    if(pos < 0)
                        pos = 0;
#ifdef DEBUG
        fprintf(stderr,"DEBUG: pos=%d, count=%d\n",pos,count);
#endif
                    if(!pos)
                        complete++;
                    else
                    {
                        ptmp1 += pos;
                        *ptmp1 = '\0';
#ifdef NOMKDIR
                        strcpy(command,"mkdir ");
#endif
                        strcpy(command,dest);
                        strcat(command,ptmp2);
#ifdef NOMKDIR
                        system(command);
#else
                        mkdir(sregp,command);
                        intcall(sregp,dregp,int21);
                        if(dregp->r_flags & F_CF)  /* if carry set */
                        {
                            erval = dregp->r_ax & 0xff;
                            if(erval == 3)  /* Path Not Found */
                            {
                                fprintf(stderr,"Can't create %s\n",command);
                                exit(4);
                            }
                            if(erval == 5)  /* Access Denied */
                                exists++;   /* directory already exists */
                        }
#endif
                        *ptmp1 = BACKSLASH;
                    }
                }
            }
            strcpy(tmp2,fullpath);
            strcpy(fullpath,dest);    
            strcat(fullpath,tmp2);

            if(overwrite)
                over++;

            exists = 0;
            if((fpout = open(fullpath,0)) > 0)
            {
                exists++;
                close(fpout);
            }

            if(!overwrite && exists)
            {                /* if overwrite not specified and file exists */
                fprintf(stderr,"\n\07%s already exists! Overwrite? ",fullpath);
                over = 0;
                i = 0;
                while(c = getchar())  /* get answer */
                {
                    if(c == 10 || c == 13)
                    {
                        ans[i] = '\0';
                        break;
                    }
                    else
                        ans[i] = c;
                    i++;
                }
                ans[0] = toupper(ans[0]); /* readkybd() should do this */
                if(strcmp(ans,"Y") == 0)
                    over++;
            }

            if(over)  /* over set from command line or previous code */
            {
                getattrib(sregp,fullpath);
                intcall(sregp,dregp,int21);
                if((dregp->r_cx & S_RDONLY) > 0)   /* check for read only */
                {
                    sregp->r_cx = dregp->r_cx ^ S_RDONLY;  /* kill that bit */
                    setattrib(sregp,fullpath,sregp->r_cx); /* change attrib */
                    intcall(sregp, dregp,int21);
                }
            }

            if(exists == 0 || over != 0)
            {
                if((fpout = creat(fullpath,2)) == MINUS1) /* -1 return */
                {
                    close(fpin);
                    fprintf(stderr,"\n%s: can't create %s\n",myname,fullpath);
                exit(3);
                }
                close(fpout);
                if((fpout = open(fullpath,1)) == MINUS1) /* -1 return */
                {
                    close(fpin);
                    fprintf(stderr,"\n%s: cannot write %s\n",myname,fullpath);
                    exit(1);
                }
                else
                {     /* successful open */
                    if(over)
                        fprintf(stderr,"\nOverwriting %s%s\n",dest,ptmp2);
                    else
                        fprintf(stderr,"\nRestoring %s%s\n",dest,ptmp2);
                    outopen++;
                }
            }
            else
                break;
        }    
    }    

      /* read in and write out using (360K floppy) track size blocks */
    while((readin = read(fpin,tmp,TRACKSIZE)) == TRACKSIZE)
    {
        write(fpout,tmp,TRACKSIZE);
    }
        /* if not 0 bytes or a full track, must be last sector(s) */
    if(readin != 0 && readin != TRACKSIZE)
    {
#ifdef SHORTFILE
        rx = readin - 1;
        if(tmp[readin] != BIGMAGIC)  /* if last word not FFFF */
#endif
            write(fpout,tmp,readin); /* write out what we read in */
#ifdef SHORTFILE
        else
        {
            write(fpout,tmp,rx);    /* write out one less byte */
        }                           /* (allow for EOF marker)  */
#endif
    }

#ifdef DEBUG
    fprintf(stderr,"last_part=%d, header[0]=%d\n",last_part,header[0]);
#endif

    if(last_part)  /* file ends on this disk */
    {
        close(fpin);

        /* set file time/date */
        putstamp(sregp,fpout,dxdate,cxtime);
        intcall(sregp,dregp,int21);
        close(fpout);

       /* set file attributes */
        setattrib(sregp, fullpath, attrib);
        intcall(sregp, dregp, int21);

        outopen = 0;
        fprintf(stderr,"\nEnter next file name or press RETURN to quit: ");
        i = 0;
        for(;;)  /* endless loop */
        {
            c = getchar();
            if(c == 10 || c ==13)
            {
                if(i == 0)
                    exit(1);
                tmpname[i] = '\0';
                break;
            }
            else
            {
                tmpname[i] = c;
                i++;
            }
        }

        if(strcmp(tmpname,"NOMORE"))
        {
            strcpy(filename,drive);
            strcat(filename,tmpname);
        }
        else
            strcpy(filename,tmpname);
    }
    else   /* file continues on next disk */
    {
        close(fpin);
        fprintf(stderr,"\n\07Insert next backup disk (%d) and press return: ",disknum+1);
        c = getchar(); /* wait for return */

        c = getbknum(backid,drive);  /* get backup disk number */
        dkc = disknum + 1;
/*        if( c % 10 == dkc % 10) */
/*            c = dkc; */

        while(c != dkc)
        {
            fprintf(stderr,"\nThe disk in source drive %c: is number %d. ",drive[0],c);
            fprintf(stderr,"It should be number %d.\n",dkc);
            fprintf(stderr,"Insert disk number %d in drive %c: and press return ('u' to use this disk): ",dkc,drive[0]);
            i = 0;
            while((c=getchar()) != '\n')
              overide[i++] = c;
            overide[i] = '\0';
/*            c = getchar(); */
/* if response is 'u', override the disk numbering */
/* skip next check and make c='next disk' value */

            if(toupper(overide[0]) == 'U')
              c = disknum + 1;
            else
              c = getbknum(backid,drive);

        }
      }
    }    
}

/* getbknum() - get number from 'backupid.@@@' file on floppy */
getbknum(name,driv)
char *name, *driv;
{
    int c, d, i;
    FILE *idfile1;

    if((idfile1 = fopen(name,"rb")) == NULL)
    {
        fprintf(stderr,"\n%s: %s unreadable or does not exist\n",myname,name);
        fprintf(stderr,"The disk in the source drive %c: does not appear to be a backup disk\n",toupper(*driv));

    c = d = i = 0;
    }
    else
    {
        for(i=0;i<2;i++)   
            c = fgetc(idfile1); /* disk number 3.2 r 1.02 and earlier */

        d = fgetc(idfile1);   /* third byte (tens digit - DOS 3.2 rel 2.02) */
        fclose(idfile1);
    }
/*    printf("c=%d    d=%d   d*10=%d   c+d=%d\n",c,d,d*10, c+d); */
    d *= 10;
    c += d;
/*    printf("c now =%d\n",c);  */
/*    exit();  */
    return(c);
}

/* usage() - prints usage message */
usage()
{
    char *vp;
    vp = sccs_id;
    vp += 5;
    fprintf(stderr,"\n%s\n",vp);
    fprintf(stderr,"\nRestores individual files from disks created with the 'backup' command\n");
    fprintf(stderr,"\nusage: %s [source_drive:]filename [-d destination_drive:] [-o] [-r]\n",myname);
    fprintf(stderr,"       default source drive is A:, default destination drive is C:\n");
    fprintf(stderr,"       -o is overwrite without asking, -r indicates source drive is ready\n");
/*    fprintf(stderr,"       destination is taken from the source file header\n"); */
}

/* register set-up routines for intcall(sregp, dregp, int#) */

/* readkybd() - unbuffered read of keyboard with no echo */
readkybd(sregp)
struct reg *sregp;
{
    sregp->r_ax = 0x08 << 8;
    sregp->r_bx = 0;
    sregp->r_cx = 0;
    sregp->r_dx = 0;
    sregp->r_si = 0;
    sregp->r_di = 0;
    sregp->r_ds = 0;
    sregp->r_es = 0;
}

/*               to duplicate timestamp on files, call getstamp, then */
/*               call putstamp with DX and CX returned by getstamp.   */

/* getstamp() -  get file date/time - AL = 0 to get                   */
/*               BX has file handle                                   */
/*               on return from intcall, DX has date, CX has time     */
getstamp(sregp, handle)
struct reg *sregp;
int handle;
{
    sregp->r_ax = 0x5700;
    sregp->r_bx = handle;
    sregp->r_cx = 0;
    sregp->r_dx = 0;
    sregp->r_si = 0;
    sregp->r_di = 0;
    sregp->r_ds = 0;
    sregp->r_es = 0;
}

/* putstamp() - set date/time of file - AL = 1 to set                    */
/*              BX = file handle, DX = date, CX = time (both packed)     */
putstamp(sregp, handle, date, time)
struct reg *sregp;
int handle;
unsigned int date, time;
{
    sregp->r_ax = 0x5701;
    sregp->r_bx = handle;
    sregp->r_cx = time;
    sregp->r_dx = date;
    sregp->r_si = 0;
    sregp->r_di = 0;
    sregp->r_ds = 0;
    sregp->r_es = 0;
}

/* getattrib() - get file attributes - DS:DX = address of ASCIIZ string */
/*               containing file's path/name - for small model, assume DS=0 */
/*               intcall() returns attributes in dregp->r_cx            */
getattrib(sregp, addr)
struct reg *sregp;
char *addr;
{
    ptoreg(dsreg,sregp->r_dx,sregp->r_ds,addr);
    sregp->r_ax = 0x4300;
    sregp->r_bx = 0;
    sregp->r_cx = 0;
    sregp->r_si = 0;
    sregp->r_di = 0;
    sregp->r_es = 0;

/* if no ptoreg() call, use the following for small model */
/*    sregp->r_ds = 0; */
/*    sregp->r_dx = addr; */ 

}


/* setattrib() - set file attributes - DS:DX = address of ASCIIZ string */
/*               containing file's path/name - for small model, assume DS=0 */
/*               intcall() sets attributes specified in CX              */
setattrib(sregp, addr, attr)
struct reg *sregp;
char *addr;
int attr;
{
    ptoreg(dsreg,sregp->r_dx,sregp->r_ds,addr);
    sregp->r_ax = 0x4301;
    sregp->r_bx = 0;
    sregp->r_cx = attr;
    sregp->r_si = 0;
    sregp->r_di = 0;
    sregp->r_es = 0;

/* if no ptoreg() call, use the following for small model */
/*    sregp->r_dx = addr; */
/*    sregp->r_ds = 0; */

}

/* mkdir() - makes a DOS directory - inputs are reg struct pointer and   */
/*           pointer to subdirectory name.  DOS returns 0 in AL for      */
/*           success, 3 for Path Not Found, 5 for Access Denied (exists) */
mkdir(sregp, addr)
struct reg *sregp;
char *addr;
{
    ptoreg(dsreg,sregp->r_dx,sregp->r_ds,addr);
    sregp->r_ax = 0x3900;
    sregp->r_bx = 0;
    sregp->r_cx = 0;
    sregp->r_si = 0;
    sregp->r_di = 0;
    sregp->r_es = 0;
}

/* get_default() - gets default drive - returns drive number in AL */
/*                 0 = A, 1 = B, 2 = C, etc.                       */
get_default(sregp)
struct reg *sregp;
{
    sregp->r_ax = 0x1900;
    sregp->r_bx = 0;
    sregp->r_cx = 0;
    sregp->r_dx = 0;
    sregp->r_si = 0;
    sregp->r_di = 0;
    sregp->r_ds = 0;
    sregp->r_es = 0;
}

findmatch(sregp,filename)
struct reg *sregp;
char *filename;
{
    ptoreg(dsreg,sregp->r_dx,sregp->r_ds,filename);
    sregp->r_ax = 0x4E00;
    sregp->r_bx = 0;
    sregp->r_cx = 7;
    sregp->r_di = 0;
    sregp->r_es = 0;
}


getdta(sregp)
struct reg *sregp;
{
    sregp->r_ax = 0x2F00;
    sregp->r_bx = 0;
    sregp->r_cx = 0;
}



/********************************
setdta(sregp,fcbp)
struct reg *sregp;
struct fcb_t *fcbp;
{
    ptoreg(dsreg,sregp->r_dx,sregp->r_ds,fcbp);
    sregp->r_ax = 0x1A00;
    sregp->r_bx = 0;
    sregp->r_cx = 0;
    sregp->r_di = 0;
    sregp->r_es = 0;
}
********************************/
